UNIT vgfit;
{Purpose: to perform least square fitting;
Version: Dec 31 1996; Mar 23 2000, May 7 2000: no anp;
         Jan 19 2001: other mem.alloc.,
         Nov 25 2001: no newton;  class: Mar 26 2002
         Jan  2 2004: constant weights
         Jan 29 2007: par records contains drv}
INTERFACE
USES vzlocom, varfun, Dialogs, SysUtils;
{ VGFIT obtaines a generalized least square minimum under constraints.
           The minimized quantity FU is:

   n                          q          km  m
  SUM w(j)*| y(xj) - f(xj,pi)|  + alpha*SUM SUM sign(gk)*gk(xj,pi)
   j                                     k   j

  Optional constraints (defining a convex set in the parameter space):

            g_l(xj,pi)   >=   0,   l = 1,...,L

  where   xj    = k-dimensional vector of independent variables,
          pi    = n-dimensional parameter vector,
          w(j) = weight of jth point, j =1,...,m
                  if w(j) = 0, jth point is ignored
          y     = experimentally measured quantity (to be fitted),
          f(x,p)= regression function,
          alpha = weight of the 2nd member (if alpha=0, only 1st member
                  of the functional FU is minimized)
          gl    = function of constraints.
          q     = 2: least square estimator
                = 1: robust estimator

     Linear dependences on free (not fixed) parameters may be imposed:

                  pk = ak * pi + bk,   k > i,    k,i IN [1..n]

  The obligatory constraints are

            p_li <= p_i <= p_ui,  i = 1,...,n

   The minimization is performed with the aid of Gauss-Newton prosess
   with the projection of the estimates onto the constraint set.
   The generalized weights in least square mode (for robustness)=

   Let be  h(j)=(y(xj)-f(xj,pi))*SQRT(w(j)). Then in each iteration
   j-th member in the minimized functional is equal to=

                   h(j)^2,           if ABS(h(j)) <= delta
              c(j)*h(j)^2,           otherwise

   where c(j)=(1+beta)/((h0(j)/delta)^2+beta),
         h0(j) is h(j) of previous iteration. }
CONST oneth:REAL = 1/1024;
      oneh:REAL = 1/128;
TYPE PARREC  = RECORD n:INTEGER; pl,p,pu,a,b,e,drv:REAL; com:STRING[11];
     {p :  at start = the initial parameter.
           at finish= the fitted  parameter.
      pl:  parameter's lower bound
      pu:  parameter's upper bound
      n :  parameter number - its position in the array; vgfit arranges
           the records.
           if n=i (i - number of parameter position in the array) the
           parameter is independent.
      a,b: if n=k and k<i, ith parameter is dependent on kth (k<i):
                       p_i = a_i * p_k + b_i
           If all the parameters are independent, n_i=i, a_i=1, b_i=0.
      e :  after fitting contains parameter errors (standard deviations)
      drv: buffer, should contain derivative with respect to parameters
      com: comment to parameter
      }
              END;
TYPE TDEB = RECORD cmn,matr,deriv,cmpdrv,cmpyf,diag,corr:BOOLEAN;
            END;
TYPE CBREC  = RECORD skip,gnit,ipr,gram,liqu,cmprs,iconv,
                     derv:INTEGER;
                     cowei:BOOLEAN;
                     dbg:TDEB;
                     eps,chi,step,damp,alpha,delta,beta:REAL;
     {Record array for the control of the minimization:
      skip  - step , with which the channels are taken during
              Gauss-Newton process (to perform the minimization
              quicklier, it is recommended to put skip=2 or 3)
      gnit  - upper limit of Gauss-Newton (GN) iterations
      eps   - the desirable accuracy of minimisation
      chi   - the desirable ratio of FU and number of degrees
              of freedom (if cb(7)=-1, it is defined automatic.)
      posd  = 1(0), the matrix should (not) be checked whether it is
                    positive definite
      step  - the step length in iterations
      damp  - contains the damping factor
      ipr   # 0 = the intermediate results of the minimization
                  (each ith set) are saved in a specified file;
              0 = no intermediate output;
      alpha = alpha for regularization
      delta = delta for the generalized (e.g., robust) fitting.
      beta  = beta   - " - " - " -    etc., see above.
      cowei = use constant weights
      gram  = 0,1
              0=full matrix minimization
              1=gradient descent minimization
      liqu  = 0 means sqrt-metric fit;
              1 means robust metric fit;
              2 means quadratic metric fit with generalized weights;
      cmprs = 1(0) - LS-matrix should (not) be compressed, i.e.
                     strings for fixed and dependent parameters
                     should be removed. Compression is used only
                     if there are no non-linear constraints on
                     parameters.
      iconv = 0,1,2;
              0=no minimization analysis while fitting
              1=if no minimization stop fitting
              2=if no minimization, exit from iteration,changing step
                   in 1,2-modes no error computation.
      derv  = 0,1,2,3;
              0=all the derivatives are computed analytically
                (by user's function ARITH)
              1=1st derivatives are computed numerically (by VGFIT)
              2=2nd derivatives are computed numerically (by VGFIT)
      dbg   = logical variables for debug output:
              cmn    - print common debug info
              matr   - print the matrix
              deriv    - print derivatives
              cmpdrv - compare numerical and analytical derivatives
              diag    - print diagonal
              corr   - print correlation factors
              cmpyf  - compare initial and fitted values,chi^2, errors
              }
              END;
TYPE RBREC  = RECORD stop,pconv,pminv,pobi,gnit,ndegf:INTEGER;
                     npfx:INTEGER; {number of fixed parameters}
                     kzer:INTEGER; {number of zero-weighted points}
                     chi2,chipl,epsf,gpoin,chir,chidev:REAL;
     {pconv = 1(0) - the process has (not) converged.
      pobi  = 1(0) - there are (no) parameters on the boundary
      ndegf = the number of degrees of freedom of FU-functional.
      epsf  = the achieved accuracy of the minimization
      gnit  = the number of used Gauss-Newton (GN) iterations
      pminv = 1(0) - the LS-matrix is (not) inverted.
      chi2  = FU achieved divided by afd
      chidev= the admissible deviation of chi2 over the unity
              IDEAL FITTING : chi2 = 1 +- chidev
      chipl = qc - ratio of the positive chi2 to the total one
              IDEAL FITTING : chipl = 0.5 +- SQRT(0.85/m)
      gpoin = part of points where chi2 is good
      chir  = the value of 2nd member in the minimized functional.
      stop  = 0 - process stopped
              1 - process was OK }
              END;
CONST gf_pnc:STRING = '++++ The process hasn''t converged';
      gf_mni:STRING = '++++ The matrix can''t be inverted';
      gf_fng:STRING = '.... The fitting quality is not good';
      gf_anr:STRING = '.... Min. accuracy not reached';
      gf_phb:STRING = '.... Parameter(s) hit the boundary:';
      gf_mxv:STRING = '.... LS matrix and vector';
      gf_miv:STRING = '.... Inverse LS matrix + vector';
      gf_nax:STRING = '.... Max (num. - anal.) deriv. = ';
      gf_drx:STRING = '.... Max der. diff. = ';
TYPE
     PROCA  = PROCEDURE(VAR y,w,f:REAL; i,typ:INTEGER);
  {A procedure of TYPE PROCA should be defined - arithmetic function
  arith for calculation of function and 1st derivivatives with respect
  to the parameters.
  If typ=1 (or 3) arith should compute and return for the point i:
  the experimental value y(x_i), weight w(x_i), the regression
  value f(x_i), the first derivatives of f with respect to the
  parameters, locating them in array drv, see examples.}
     PROCR  = PROCEDURE(iit:INTEGER);
  {A procedure of TYPE PROCR should be defined which is regulizer
   or projector; it must calculate value of the second member of FU and
   its derivatives with respect to parameters and can correct
   (currently), if needed, a, b, mat, yar etc., or display some
   intermediate results, as, e.g., current chi^2 etc; see examples.}
TYPE TParameter = CLASS
      PUBLIC { Public declarations }
  compstr:STRING;       {info about compression}
  fixdepstr:STRING;     {info about fixing and depend.controv.}
  freepstr:STRING;      {info if no free par.}
  parnumstr:STRING;     {info if parameter numeration is correct}
  cb:CBREC;
  rb:RBREC;
  par:ARRAY OF PARREC;  {parameter record array}
  npar:INTEGER;         {N of initial parameters: both fixed and free}
  nmat:INTEGER;         {N of compressed parameters}
  mpoint:INTEGER;       {Number of points (x_j), j=1,2,...,mpoint}
  bayes:BOOLEAN;        {Bayesian or not pick-up of points}
  nocnv,noinv,noch,noeps,noequ:BOOLEAN; {convergence checks}
  ww:REAL;              {Value from compressed matrix}
  v_stop:BOOLEAN;       {the process should be stopped}
  yar:ARRAY OF REAL;    {Array for LS-vector}
  err_av:REAL;          {average error of exp.data}
  FCorr:REAL;           {Global correlation factor}
  tergnit:INTEGER; {normally gnit; at last trial 3*gnit}
  terstep:REAL;    {normally step; at last trial step/2}
  tertria:BOOLEAN; {false=current trial; true=terminating}
  typ:INTEGER;
  ifail:INTEGER;        {0=matrix not inverted}
  nocomp,noldep:BOOLEAN;{is compression?}
  chisr,chiold,chinew:REAL; {Chi-squares, current and foregoing}
  cb0:CBREC;
  CONSTRUCTOR Create(np,nm,m:INTEGER);
  PROCEDURE readm(i,j:INTEGER);
  PROCEDURE writem(i,j:INTEGER);
  PROCEDURE FormCB;
  PROCEDURE rincc;
  PROCEDURE FormParameters;
  PROCEDURE CheckConvergence;
  PROCEDURE CheckLinDepend(VAR nmat:INTEGER);
  PROCEDURE GetNFreePar(VAR npfr:INTEGER);
  PROCEDURE LeastSquaresFit(arith:PROCA; regul:PROCR);
  PROCEDURE podem;
  PROCEDURE PrintMatr(CONST s:STRING);
  PROCEDURE prifch;
  PROCEDURE lsmprn(aria:PROCA; VAR cmax:REAL);
  PROCEDURE primat;
  PROCEDURE Arith(VAR y,w,f:REAL; i,typ:INTEGER); VIRTUAL;
  PROCEDURE Regul(iit:INTEGER); VIRTUAL;
  PROCEDURE Prepare00;
  PROCEDURE Check00(VAR ok:BOOLEAN);
  PROCEDURE ZeroVecMat;
  PROCEDURE FormMatVect(rr,rrr,w:REAL);
  PROCEDURE CompleteMatr(VAR grad:REAL);
  PROCEDURE DampMatrix;
  PROCEDURE InvSMatrix;
  PROCEDURE CorrectPars(mo:INTEGER);
  PROCEDURE ExpandVect;
  PROCEDURE MakeAccu(grad:REAL);
  PROCEDURE SaveErrors;
  PROCEDURE PrintParArray;
      PRIVATE { Private declarations }
  mat,dia,ndrv:ARRAY OF REAL;
                        {Matrix, diagonal,vector,numeric derivatives}
  {fout  = textfile variable for output, defined in vzlocom
   The fitted parameter values are in the array par (par.p)
   The covariance matrix is in the array mat.
   The parameter errors are in the array par (par.e).
   The correlation factors are in the array yar.}
  text:ARRAY[1..40] OF CHAR;
  PROCEDURE outin(typ:INTEGER);
  PROCEDURE Minimize(VAR i0:INTEGER; arith:PROCA; regul:PROCR);
  PROCEDURE fixmat(VAR ks:INTEGER);
  PROCEDURE accldp(VAR ks:INTEGER);
  PROCEDURE CompareDeriv(VAR derx:REAL; VAR idx:INTEGER);
  PROCEDURE NumericDiff(kp:INTEGER; f00:REAL; arith:PROCA);
  END;


PROCEDURE TestVGFit;
PROCEDURE TestRegul(iit:INTEGER);

CONST blank:CHAR =' '; plus:CHAR ='+'; astr:CHAR ='*';

VAR    testlsm:TParameter;
{-----------------------------------------------}
IMPLEMENTATION
PROCEDURE TParameter.LeastSquaresFit(arith:PROCA; regul:PROCR);
{ Version 19 December 1989 }
VAR   i0,i:INTEGER; ok:BOOLEAN;
LABEL L3;
{-----------------------------------------------------------}
BEGIN Prepare00; Check00(ok); IF(NOT ok)THEN Exit; v_stop:=false;
{Gauss-Newton process}
terstep:=cb.step; i0:=1; typ:=1; IF(cb.gnit=0)THEN GOTO L3;
REPEAT rb.pconv:=1; Minimize(i0,arith,regul);
ConvertStr(chisr,12,4); IF(cb.dbg.cmn)THEN
 ShowMess(IntToStr(rb.gnit)+' '+sedz+'-'+IntToStr(rb.ndegf));
Inc(rb.gnit); IF((rb.stop<>1)OR(rb.pminv=0))THEN GOTO L3;
UNTIL(
 (rb.gnit>=cb.gnit)OR((rb.epsf<=cb.eps)AND(chisr<=cb.chi))OR(v_stop));
{the output data calculation}
L3: typ:=3; IF(cb.ipr<>0)THEN outin(typ); cb.skip:=1;
FOR i:=1 TO npar DO par[i].e:=0;
IF(rb.pminv<>0)THEN Minimize(i0,arith,regul); cb:=cb0;
END;
PROCEDURE TParameter.Minimize(VAR i0:INTEGER; arith:PROCA; regul:PROCR);
VAR   kp,mg,idx:INTEGER;
      y,w,f,h,h2,rr,wrob,rrr,grad,chi,qc,wsq,derx:REAL;
LABEL L125,L23;
BEGIN ZeroVecMat; chi:=0; qc:=0; mg:=0;
derx:=0; kp:=i0;
REPEAT arith(y,w,f,kp,typ); IF(w=0)THEN{loop in points}
 BEGIN Inc(rb.kzer); GOTO L125;
 END;
IF(rb.gnit=0)THEN err_av:=err_av+1/SQRT(ABS(w));
WITH rb DO
 IF((gnit>0)AND(cb.cowei)AND(typ<>3))THEN w:=SQR(ndegf/err_av);
IF(cb.derv<>0)THEN NumericDiff(kp,f,arith); {Numer.diff.}
IF(cb.dbg.cmpdrv)THEN CompareDeriv(derx,idx);{Comp.num.& analyt.deriv.}
rr:=y-f; wsq:=SQRT(w); h:=ABS(rr)*wsq;{Process the weights}
CASE cb.liqu OF {Choose metric: SQRT,modulus,quadratic?}
0:{sqrt}
  BEGIN w:=SQRT(wsq); w:=w/AMAX1(h*SQRT(ABS(h)),0.000001);
  h2:=SQR(ABS(rr))*w;
  END;
1:{modul}
  BEGIN w:=wsq/AMAX1(ABS(rr),0.000001); h2:=SQR(rr)*w;
  END;
2:{quadr}
  BEGIN h2:=h*h; IF(h>=cb.delta)THEN
   BEGIN wrob:=(1+cb.beta)/(SQR(h/cb.delta)+cb.beta);
   w:=w*wrob; h2:=h2*wrob;
   END;
  END;
ELSE END;
chi:=chi+h2; IF(h2<(1+SQRT(18/mpoint)))THEN mg:=mg+1;
IF(rr>0)THEN qc:=qc+h2; rrr:=rr*w;
FormMatVect(rr,rrr,w);
IF((cb.dbg.cmpyf)AND(typ=3))THEN
 Writeln(fout,kp:5,y:10:2,f:10:2,h2:10:2,(1/SQRT(w)):10:2);
L125: kp:=kp+cb.skip;
UNTIL(kp>mpoint); {end loop in points}
IF(rb.kzer=mpoint)THEN
 BEGIN ShowMess('All errors=0.Fit stopped'); Abort;
 END;
CompleteMatr(grad);
rb.chi2:=chi/rb.ndegf; rb.chipl:=qc/AMAX1(0.000001,chi);;
rb.gpoin:=mg/MAX0(1,mpoint);
regul(kp);{the regularisation,if required}
DampMatrix;
InvSMatrix; IF(ifail=0)THEN {the matrix inversion}
 BEGIN IF(typ<>3)THEN rb.pconv:=0; rb.pminv:=0; Exit;
 END;
IF((typ=3)AND(nocomp))THEN podem;{is the reciprocal matrix pos.defin.}
IF(typ<>3)THEN CorrectPars(0);
ExpandVect;
SaveErrors;
IF(typ=3)THEN Exit; {save the errors}
chisr:=rb.chi2+cb.alpha*rb.chir;
IF(cb.iconv=0)THEN GOTO L23; {stop if inconvergence}
IF((cb.iconv=1)AND(chisr>chiold))THEN
 BEGIN rb.stop:=0; Exit;
 END;
IF(cb.iconv=2)THEN
 BEGIN IF(chisr<=chiold)THEN GOTO L23; WITH cb DO
  BEGIN step:=step*0.5; damp:=AMIN1(damp*2,100); skip:=1; Exit;
  END;
 END;
L23: chiold:=chisr;
CorrectPars(1);
MakeAccu(grad);
IF(i0=cb.skip)THEN i0:=0; Inc(i0);
END;
PROCEDURE TParameter.MakeAccu(grad:REAL);
VAR i:INTEGER; x:REAL;
BEGIN x:=0; FOR i:=1 TO npar DO WITH par[i] DO {compute the accuracy}
 IF((n=i)AND(dia[i]<>0))THEN x:=x+SQR(p-drv)*dia[i];
x:=SQRT(ABS(x)/nmat); rb.epsf:=SQRT(x)*SQRT(grad);
IF(cb.dbg.diag)THEN
 Writeln(fout,'Diag= ',grad:15:4,x:10:4,rb.epsf:15:4,rb.chi2:10:3);
END;
PROCEDURE TParameter.ExpandVect;
VAR i,k:INTEGER;
BEGIN IF(nocomp)THEN FOR i:=1 TO npar DO dia[i]:=mat[(i-1)*npar+i]
                ELSE {expand the vector}
 BEGIN FOR i:=1 TO npar DO yar[i]:=par[i].drv; k:=1;
 FOR i:=1 TO npar DO WITH par[i] DO
 IF((n<>i)OR(pl=pu))THEN
  BEGIN drv:=0; dia[i]:=0;
  END               ELSE
   BEGIN drv:=yar[k]; readm(k,k); dia[i]:=ww; Inc(k);
   END;
 END;
END;
PROCEDURE TParameter.CorrectPars(mo:INTEGER);
VAR   i,j:INTEGER; xl,xu,rr:REAL;
LABEL L2;
BEGIN IF(mo=1)THEN GOTO L2;
IF((cb.ipr<>0)AND((rb.gnit MOD cb.ipr)=0))THEN outin(typ);{print}
{calculate parameter corrections and norm of vector}
FOR i:=1 TO nmat DO WITH par[i] DO
 BEGIN drv:=0; FOR j:=1 TO nmat DO
  BEGIN readm(i,j); drv:=drv+ww*yar[j];
  END;
 END;
IF(cb.dbg.matr)THEN PrintMatr(gf_miv);
FCorr:=0; FOR i:=1 TO nmat DO {Correlation factor}
 BEGIN readm(i,i); FCorr:=FCorr+ww*dia[i];
 END;
FCorr:=FCorr/nmat;
IF(cb.dbg.corr)THEN Writeln(fout,'FofCorr= ',FCorr:10:2); Exit;
L2: FOR i:=1 TO npar DO WITH par[i] DO {correct the parameter values}
 BEGIN rr:=p+terstep*drv; drv:=p; p:=rr;
 END;
rb.pobi:=0; FOR i:=1 TO npar DO WITH par[i] DO {project onto par.bounds}
 BEGIN rr:=p; xl:=pl; xu:=pu;
 IF(rr<xl)THEN p:=xl; IF(rr>xu)THEN p:=xu;
 IF((pl<>pu)AND(n=i)AND((p=pl)OR(p=pu)))THEN rb.pobi:=1;
 END;
IF(NOT noldep)THEN FOR i:=1 TO npar DO{correct for linear dependences}
 IF(par[i].n<>i)THEN WITH par[i] DO
  BEGIN k:=Round(n); p:=a*par[k].p+b;
  END;
END;
PROCEDURE TParameter.DampMatrix;
VAR i:INTEGER; rr:REAL;
BEGIN IF((typ=3)AND(nocomp))THEN{store the diag}
 FOR i:=1 TO nmat DO yar[i]:=mat[(i-1)*nmat+i];
IF(cb.damp<>0)THEN FOR i:=1 TO nmat DO{Damping}
 IF(NOT((nocomp)AND(par[i].pl=par[i].pu)))THEN
 BEGIN readm(i,i); IF(typ<>3)THEN rr:=cb.damp ELSE rr:=0.00001;
 ww:=ABS(ww)*(1+rr); IF(ABS(ww)<1E-7)THEN ww:=rr; writem(i,i);
 END;
IF((typ=3)AND(nocomp))THEN podem; {is the matrix positive definite}
END;
PROCEDURE TParameter.CompleteMatr(VAR grad:REAL);
VAR i,j:INTEGER;
BEGIN FOR i:=1 TO nmat DO{complete the matrix}
 BEGIN FOR j:=i TO nmat DO
  BEGIN readm(i,j); writem(j,i);
  END;
 ww:=dia[i]; writem(i,i);
 END;
grad:=0; FOR i:=1 TO nmat DO grad:=grad+SQR(yar[i]);
grad:=SQRT(grad/nmat);
rb.npfx:=0; IF(nocomp)THEN{fixation  (if no compression)}
 BEGIN fixmat(rb.npfx);{account for linear depend.if any (if no compr.)}
 IF(NOT noldep)THEN accldp(rb.npfx);
 END;
rb.ndegf:=(mpoint-rb.kzer); rb.ndegf:=rb.ndegf DIV cb.skip-nmat+rb.npfx;
IF(rb.ndegf<=0)THEN rb.ndegf:=1; rb.chidev:=SQRT(2/rb.ndegf);
IF(cb.chi<0)THEN cb.chi:=1+rb.chidev;
END;
PROCEDURE TParameter.FormMatVect(rr,rrr,w:REAL);
VAR i,j,kk:INTEGER; rd:REAL;
BEGIN IF(NOT nocomp)THEN{compress the drv-vector}
 BEGIN FOR i:=1 TO npar DO WITH par[i] DO
  BEGIN j:=Round(n);
  IF(j<>i)THEN par[j].drv:=par[j].drv+drv*a;
  END;
 kk:=1; FOR i:=1 TO npar DO WITH par[i] DO IF(pl<>pu)THEN IF(n=i)THEN
  BEGIN par[kk].drv:=drv; Inc(kk);
  END;
 END;

IF(cb.dbg.deriv)THEN FOR j:=1 TO nmat DO {Print derivatives}
  BEGIN Write(fout,par[j].drv:13:4);
  IF(((j MOD 5)=0)OR(j=nmat))THEN Writeln(fout,'  deriv');
  END;

FOR i:=1 TO nmat DO{form vector and matrix}
 BEGIN yar[i]:=yar[i]+rrr*par[i].drv; FOR j:=i TO nmat DO
  BEGIN rd:=par[i].drv*par[j].drv;
  IF((typ=3)AND(nocomp))THEN
   rd:=rd-rr*mat[(i-1)*nmat+j]; rd:=rd*w;
  IF(i=j)THEN dia[i]:=dia[i]+rd
         ELSE
   BEGIN readm(i,j); ww:=ww+rd; writem(i,j);
   END;
  END;
 END;
IF(cb.dbg.diag)THEN FOR j:=1 TO nmat DO  {Print diagonal}
  BEGIN Write(fout,dia[j]:14:4);
  IF(((j MOD 5)=0)OR(j=nmat))THEN Writeln(fout);
  END;
END;
PROCEDURE TParameter.NumericDiff(kp:INTEGER; f00:REAL; arith:PROCA);
VAR i,j:INTEGER;
    derv,x,f,ph,stp,phi,phj,f10,f01,f11,fp1,fm1:REAL;
BEGIN IF((cb.derv=1)OR(cb.dbg.cmpdrv))THEN {1st derivatives}
 BEGIN FOR i:=1 TO npar DO
  BEGIN ndrv[i]:=0; IF(par[i].pl=par[i].pu)THEN CONTINUE;
  ph:=par[i].p; stp:=ph/oneth; par[i].p:=ph+stp; arith(x,x,f,kp,typ);
  derv:=(f-f00)/AMAX1(stp,1E-6); ndrv[i]:=derv; par[i].p:=ph;
  END;
 FOR i:=1 TO npar DO par[i].drv:=ndrv[i];
 END;
IF((typ<>3)OR(NOT nocomp)OR(npar>10))THEN Exit;
FOR i:=1 TO npar DO{2nd derivatives}
 BEGIN phi:=par[i].p; FOR j:=i TO npar DO
 IF((par[i].pl=par[j].pu))THEN
  BEGIN ww:=0; writem(j,i);
  END                       ELSE
  BEGIN phj:=par[j].p;
  IF(i<>j)THEN {Different i,j}
   BEGIN par[i].p:=phi+oneh; par[j].p:=phj;
   arith(x,x,f,kp,typ); {x+h,y}
   f10:=f; par[i].p:=phi; par[j].p:=phj+oneh;
   arith(x,x,f,kp,typ); {x,y+h}
   f01:=f; par[i].p:=phi+oneh; par[j].p:=phj+oneh;
   arith(x,x,f,kp,typ); {x+h,y+h}
   f11:=f; derv:=(f11+f00-f01-f10)*128*128;
   END    ELSE {Equal i,j}
   BEGIN par[i].p:=phi+oneh; arith(x,x,f,kp,typ); {x+h}
   fp1:=f; par[i].p:=phi-oneh; arith(x,x,f,kp,typ); {x-h}
   fm1:=f; derv:=(fp1-2*f00+fm1)*128*128;
   END;
  ww:=derv; writem(j,i); par[i].p:=phi; par[j].p:=phj;
  END;
 END;
END;
PROCEDURE TParameter.Check00(VAR ok:BOOLEAN);
VAR i:INTEGER; wr:BOOLEAN;
BEGIN ok:=true; IF((mpoint<=0)OR(npar<=0))THEN
 BEGIN ShowMess('No points (or no parameters)'); ok:=false;
 END;
FOR i:=1 TO npar DO WITH par[i] DO
 BEGIN wr:=(n>i)OR(n=0)OR(a=0); IF(wr)THEN {Check parameter array}
  BEGIN parnumstr:='Wrong parameter array'; rb.stop:=0; ok:=false;
  ShowMess(parnumstr);
  END;
 END;
IF(nmat=0)THEN {Check parameter dependences and compression}
 BEGIN freepstr:='Error, no free parameters'; ok:=false;
 END;
END;
PROCEDURE TParameter.ZeroVecMat;
VAR i,j:INTEGER;
BEGIN rb.kzer:=0; FOR i:=1 TO npar DO
 BEGIN yar[i]:=0; dia[i]:=0; par[i].drv:=0;
 END;
ww:=0; FOR i:=1 TO nmat DO FOR j:=1 TO nmat DO writem(j,i);
END;
PROCEDURE TParameter.Prepare00;
BEGIN cb0:=cb; cb.skip:=MAX0(cb.skip,1); IF(cb.delta=0)THEN cb.delta:=1;
rb.pminv:=1; rb.gnit:=0; rb.epsf:=1.E16; err_av:=0;
rb.chir:=0; rb.stop:=1; chiold:=1.E16;
nocomp:=(cb.cmprs=0);
compstr:=''; fixdepstr:='';  freepstr:=''; parnumstr:='';
CheckLinDepend(nmat);
END;
PROCEDURE TParameter.fixmat(VAR ks:INTEGER);
{Purpose: To fix the parameters; Version: October 29 1992}
VAR   i,k:INTEGER;
LABEL L1;
BEGIN IF(cb.gram=0)THEN GOTO L1;
FOR i:=1 TO npar DO IF(par[i].pl=par[i].pu)THEN
  BEGIN yar[i]:=0; Inc(ks);
  END;
Exit;
L1: FOR i:=1 TO npar DO IF(par[i].pl=par[i].pu)THEN
 BEGIN FOR k:=1 TO npar DO
  BEGIN mat[(k-1)*npar+i]:=0; mat[(i-1)*npar+k]:=0;
  END;
 ww:=1; writem(i,i); yar[i]:=0; Inc(ks);
 END;
END;
PROCEDURE TParameter.accldp(VAR ks:INTEGER);
{Purpose: To account for linear dependencies; Version: October 29 1992}
VAR   i,j,k,l,n:INTEGER;
LABEL L1;
BEGIN n:=npar; IF(cb.gram=0)THEN GOTO L1;
FOR i:=1 TO n DO IF(par[i].n<>i)THEN
 BEGIN j:=par[i].n; yar[j]:=yar[j]+par[i].a*yar[i];
 yar[i]:=0; Inc(ks);
 END;
Exit;
L1: FOR i:=1 TO n DO IF(par[i].n<>i)THEN
 BEGIN j:=par[i].n; FOR k:=1 TO n DO
  BEGIN mat[(j-1)*n+k]:=mat[(j-1)*n+k]+mat[(i-1)*n+k]*par[i].a;
  mat[(k-1)*n+j]:=mat[(k-1)*n+j]+mat[(k-1)*n+i]*par[i].a;
  END;
 mat[(j-1)*n+j]:=mat[(j-1)*n+j]+par[i].a*par[i].a*mat[(i-1)*n+i];
 yar[j]:=yar[j]+par[i].a*yar[i];
 FOR k:=1 TO n DO
  BEGIN ww:=0; writem(i,k); writem(k,i);
  END;
 yar[i]:=0; ww:=1; writem(i,i); Inc(ks);
 END;
END;
PROCEDURE TParameter.InvSMatrix;
{Purpose: inverts a symmetrical matrix  a(ndim,ndim);
          if failure then ifail:=1 else ifail:=0;  Version: 2.11.1988}
VAR   i,j,k,l,np,np1,jj,kk,ij,kj,ji,jl,ij0,ji0:INTEGER; elm,pivot:REAL;
LABEL L100,L19;
BEGIN ifail:=1; np1:=nmat+1; FOR i:=1 TO nmat DO par[i].e:=1;
FOR i:=1 TO nmat DO
 BEGIN {Find pivot} pivot:=0; jj:=1; FOR j:=1 TO nmat DO
  BEGIN IF(par[j].e=0)THEN GOTO L19; elm:=ABS(mat[jj]);
  IF(elm<=pivot)THEN GOTO L19; pivot:=elm; k:=j; kk:=jj;
  L19: jj:=jj+np1;
  END;
 IF(pivot<1.E-16)THEN GOTO L100;
 par[k].e:=0; pivot:=-mat[kk];
 {Eliminate}
 kj:=k; np:=nmat; FOR j:=1 TO nmat DO
  BEGIN IF((j-k)=0)THEN
   BEGIN mat[kj]:=1/pivot; par[j].drv:=0; np:=1;
   END             ELSE
    BEGIN elm:=-mat[kj]; par[j].drv:=elm/pivot;
    IF(elm<>0)THEN
     BEGIN jl:=j; FOR l:=1 TO j DO
      BEGIN mat[jl]:=mat[jl]+elm*par[l].drv; jl:=jl+nmat;
      END;
     END;
    mat[kj]:=par[j].drv;
    END;
  kj:=kj+np;
  END;
 END;
{Change the sign and fill up provisionally}
ij0:=1; ji0:=1; FOR i:=1 TO nmat DO
 BEGIN ij:=ij0; ji:=ji0; FOR j:=1 TO i DO
  BEGIN mat[ij]:=-mat[ij]; mat[ji]:=mat[ij]; ij:=ij+nmat; Inc(ji);
  END;
 ij0:=ij0+1; ji0:=ji0+nmat;
 END;
Exit;
L100: ifail:=0; {Failure return}
END;
PROCEDURE TParameter.podem;
{Purpose: To check whether matrix is positive definite;  Version: 1976}
VAR   i,j,k:INTEGER; rr,qi,qj,d:REAL;
LABEL L128;
BEGIN rb.pconv:=1; FOR i:=1 TO nmat DO
 IF(mat[(i-1)*nmat+i]<0)THEN GOTO L128; IF(nmat<2)THEN Exit;
k:=nmat-1; FOR i:=1 TO k DO
 BEGIN FOR j:=i+1 TO nmat DO
  BEGIN rr:=mat[(i-1)*nmat+j]; qi:=mat[(i-1)*nmat+i];
  qj:=mat[(j-1)*nmat+j]; d:=AMAX3(ABS(rr),qi,qj);
  IF(d=0)THEN GOTO L128;
  rr:=(qi/d)*qj-(rr/d)*rr; IF(rr<=0)THEN GOTO L128;
  END;
 END;
Exit;
L128: rb.pconv:=0;
END;
PROCEDURE TParameter.outin(typ:INTEGER);
{Purpose: The subroutine outputs intermediate results of the;
          Least-Square minimization.  Version 19.12.1989}
VAR i,j,k,l,ni:INTEGER; {hour,min,sec,is:WORD;}
BEGIN ni:=rb.gnit+1; {GetTime(hour,min,sec,is);}
IF(cb.ipr<>0)THEN
 BEGIN {Writeln(fv,'Time=',hour:2,'.',min:2,'.',sec:2,'  ',
                  '  Ni     Chi2/df       eps       step      damp');}
 Writeln(fout,'Iter.= ',ni:4,'  chi2= ',rb.chi2:8:2,'   Eps= ',
 rb.epsf:5:1,' Step= ',cb.step:6:3,' Damp= ',cb.damp:9:6);
 k:=npar DIV 5; l:=npar-k*5; FOR i:=1 TO k DO
  BEGIN FOR j:=1 TO 5 DO Write(fout,par[(i-1)*5+j].p:12:4);
  Writeln(fout);
  END;
 IF(l<>0)THEN
  BEGIN FOR j:=1 TO l DO Write(fout,par[k*5+j].p:12:4);
  Writeln(fout);
  END;
 END;
END;
PROCEDURE TParameter.CheckLinDepend(VAR nmat:INTEGER);
{Purpose: To check parameter dependences and compression;
 Version: October 13 1992}
VAR   i:INTEGER; st1,st2:STRING; r:REAL;
BEGIN
{Check whether there are linear dependences}
noldep:=true; FOR i:=1 TO npar DO WITH par[i] DO IF(n<>i)THEN
 BEGIN noldep:=false; IF(n<i)THEN
  BEGIN pl:=a*par[n].pl+b; p:=a*par[n].p+b; pu:=a*par[n].pu+b;
  Arrange2(pl,pu);
  END;
 END;
FOR i:=1 TO npar DO WITH par[i] DO
 BEGIN Arrange2(pl,pu); IF((n<>i)AND(pl=pu))THEN
  BEGIN n:=i;
  fixdepstr:='Warning! Fixing and dependence contradiction removed';
  END;
 END;
{build the real dimension of the LS-matrix for compression}
nmat:=npar; IF(nocomp)THEN Exit;
FOR i:=1 TO npar DO
 IF(par[i].pl=par[i].pu)THEN Dec(nmat);
IF(NOT noldep)THEN FOR i:=1 TO npar DO IF(par[i].n<>i)THEN Dec(nmat);
r:=npar; ConvertStr(r,4,0); st1:=sedz;
r:=nmat; ConvertStr(r,4,0); st2:=sedz;
compstr:='.... Compression: from '+st1+'  to '+st2;
END;
PROCEDURE TParameter.GetNFreePar(VAR npfr:INTEGER);
{Get the number of free parameters
 Version: October 13 1992;  Jan 8 2003}
VAR i:INTEGER;
BEGIN npfr:=npar;
FOR i:=1 TO npar DO IF(par[i].pl=par[i].pu)THEN Dec(npfr);
FOR i:=1 TO npar DO IF(par[i].n<>i)THEN Dec(npfr);
END;
PROCEDURE TParameter.readm(i,j:INTEGER);
{Purpose: to read a matrix mat(npar,npar) item (i,j)  into ww
 Version: 21.12.1988}
BEGIN ww:=mat[(j-1)*nmat+i];
END;
PROCEDURE TParameter.writem(i,j:INTEGER);
{Purpose: to write ww into a matrix mat(npar,npar) item (i,j);
 Version: 21.12.1988}
BEGIN mat[(j-1)*nmat+i]:=ww;
END;
PROCEDURE TParameter.rincc;
{varies minimisation control data, if inconvergence; version 86.01.20}
BEGIN cb.damp:=AMIN1(cb.damp*10,0.1); cb.step:=cb.step*0.5;
cb.skip:=1; IF(cb.gnit<>0)THEN cb.gnit:=cb.gnit+1;
IF((cb.gnit>1)AND(cb.gnit<=4))THEN cb.gnit:=5;
END;
PROCEDURE TParameter.FormParameters;
VAR i:INTEGER;
BEGIN FOR i:=1 TO npar DO WITH par[i] DO
 BEGIN n:=i; a:=1; b:=0; com:=''; pl:=-1E15; p:=0; pu:=1E15;
 e:=0; drv:=0;
 END;
END;
PROCEDURE TParameter.CheckConvergence;
BEGIN nocnv:=rb.pconv<>1; noinv:=rb.pminv<>1; noch:=rb.chi2>3;
noeps:=rb.epsf>cb.eps;
noequ:=((rb.chipl<0.25)OR(rb.chipl>0.75))AND(rb.chi2>0.1);
END;
PROCEDURE TParameter.FormCB;
{ Purpose: To form the standard cb and iar,aar,bar blocks
  Version: Jun 28 1998, Mar 24 2000}
VAR i:INTEGER;
BEGIN WITH cb DO
 BEGIN skip:=1; gnit:=5; ipr:=0; gram:=0; cowei:=false;
 liqu:=2; cmprs:=0; iconv:=0; derv:=0;
 eps:=0.5; chi:=1; step:=0.999; damp:=0.00001; alpha:=0;
 delta:=1000000; beta:=1000000; WITH dbg DO
  BEGIN cmn:=false; matr:=false; deriv:=false; cmpdrv:=false;
  cmpyf:=false; diag:=false; corr:=false;
  END;
 END;
bayes:=false;
END;
PROCEDURE TParameter.PrintMatr(CONST s:STRING);
VAR i,j:INTEGER; nrm:REAL;
BEGIN nrm:=0; FOR i:=1 TO nmat*nmat DO nrm:=AMAX1(nrm,ABS(mat[i]));
nrm:=AMAX1(nrm,1); Writeln(fout,s,'  norm.c.= ',nrm:15:7);
FOR i:=1 TO nmat DO FOR j:=1 TO nmat DO
 BEGIN readm(i,j); Write(fout,ww/nrm:7:4);
 IF(((j MOD 10)=0)OR(j=nmat))THEN Writeln(fout);
 END;
Writeln(fout); FOR j:=1 TO nmat DO
 BEGIN Write(fout,yar[j]/nrm:7:3);
 IF(((j MOD 10)=0)OR(j=nmat))THEN Writeln(fout);
 END;
END;
PROCEDURE TParameter.SaveErrors;
VAR i,k:INTEGER;
BEGIN FOR i:=1 TO npar DO
 IF(par[i].pl<>par[i].pu)THEN par[i].e:=SQRT(ABS(dia[i]))
                         ELSE par[i].e:=0;
IF(NOT noldep)THEN FOR i:=1 TO npar DO IF(par[i].n<>i)THEN
 BEGIN k:=par[i].n; IF(k<>i)THEN par[i].e:=ABS(par[i].a*par[k].e);
 END;
END;
PROCEDURE TParameter.CompareDeriv(VAR derx:REAL; VAR idx:INTEGER);
VAR i,j:INTEGER; rr,rrr:REAL;
BEGIN rr:=0; j:=1; FOR i:=1 TO npar DO WITH par[i] DO
 BEGIN rrr:=ABS(drv-ndrv[i]);
 IF((pl<>pu)AND(rrr>rr))THEN
  BEGIN rr:=rrr; j:=i;
  END;
 END;
{Writeln(fout,gf_nax,drv[j]:10:3,ndrv[j]:10:3,j:5);}
IF(rr>derx)THEN
 BEGIN derx:=rr; idx:=j;
 END;
IF(cb.dbg.deriv)THEN Writeln(fout,gf_drx,derx:10:3,idx:5);
END;
PROCEDURE TParameter.prifch;
{Purpose: To print out the fitting numerical characteristics and
          check for the fit quality; Version: 21.11.1988, Jun 28 1998}
VAR i:INTEGER; log:BOOLEAN;
BEGIN CheckConvergence;
IF(parnumstr<>'')THEN Writeln(fout,parnumstr);
IF(freepstr<>'')THEN Writeln(fout,freepstr);
IF(compstr<>'')THEN Writeln(fout,compstr);
IF(fixdepstr<>'')THEN Writeln(fout,fixdepstr);
IF(rb.stop<>1)THEN Exit;
IF(nocnv)THEN Writeln(fout,gf_pnc);
IF(noinv)THEN Writeln(fout,gf_mni );
IF((noch)OR(noequ))THEN Writeln(fout,gf_fng);
IF(noeps)  THEN Writeln(fout,gf_anr);
IF(rb.pobi=1)THEN Writeln(fout,gf_phb );
FOR i:=1 TO npar DO WITH par[i] DO
 BEGIN log:=((p=pl)OR(p=pu)AND(pl<>pu));
 IF(log)THEN Writeln(fout,i:5,p:10:3);
 END;
Writeln(fout,' Ch2/df=',rb.chi2:7:2,'  eps= ',rb.epsf:5:1,'   ',
rb.gnit:3,' iterations','  Ch+=',rb.chipl:4:2);
Writeln(fout,' ':20,'Parameters and their errors');
FOR i:=1 TO npar DO Writeln(fout,par[i].p:15:3,'   +-',par[i].e:15:3);
END;
PROCEDURE TParameter.lsmprn(aria:PROCA; VAR cmax:REAL);
{Purpose: output the fitting results and build the fitted curve
          Version: June 30 1993, Jun 17 1998}
VAR  i,j,nd,ik:INTEGER; c,cmin,step,y,w,v:REAL;
  FUNCTION kkk(x:REAL):INTEGER;
  BEGIN kkk:=MAX0(MIN0(Round((x-cmin)/step+0.5)+1,40),1)
  END;
{-----------------------------------------------}
BEGIN cmin:=0;
Writeln(fout,'  Data(+)','  Fit(*)','      Chi2','             Graph');
IF(cmax=0)THEN {Find max,min for scale}
 BEGIN cmax:=-1E15; cmin:=1E15; FOR i:=1 TO mpoint DO
  BEGIN aria(y,w,v,i,3); cmax:=AMAX1(cmax,y); cmin:=AMIN1(cmin,y);
  END;
 END;
cmin:=AMIN1(0,cmin); nd:=Round(LN(cmax-cmin+1)*0.434);
CASE nd OF
0: nd:=4;
1: nd:=3;
2: nd:=2;
ELSE nd:=1; END;
{Scale and the main loop}
cmax:=cmax+SQRT(ABS(cmax)); step:=AMAX1((cmax-cmin)/40,0.001);
FOR i:=1 TO mpoint DO
 BEGIN FOR j:=1 TO 40 DO text[j]:=blank;
 aria(y,w,v,i,3); c:=(y-v)*(y-v)*w;
 ik:=kkk(y); text[ik]:=plus; ik:=kkk(v); text[ik]:=astr;
 Write(fout,y:9:nd,v:9:nd,c:9:nd);
 FOR j:=1 TO 40 DO Write(fout,text[j]); Writeln(fout);
 END;
END;
PROCEDURE TParameter.PrintParArray;
VAR i:INTEGER; s:STRING;
BEGIN Writeln(fout,
 '   n      par_low         par          par_up      a     b');
FOR i:=1 TO npar DO WITH par[i] DO
 BEGIN ConvertStr(Float(n),5,0); s:=sedz; ftos(pl,14,7,sedz); s:=s+sedz;
 ftos(p,14,7,sedz); s:=s+sedz; ftos(pu,14,7,sedz); s:=s+sedz;
 ConvertStr(a,6,3); s:=s+sedz; ConvertStr(b,6,3); s:=s+sedz;
 s:=s+com; Writeln(fout,s);
 END;
END;
PROCEDURE TParameter.primat;
{Purpose: To output the correlation matrix; Version: Feb 25 1998;}
VAR i,j:INTEGER; cmax,spu,rr:REAL;
BEGIN cmax:=0; spu:=0; IF(cb.cmprs<>0)THEN
 BEGIN Writeln(fout,'Invalid operation for compressed matrix'); Exit;
 END;
IF(rb.stop<>1)THEN Exit;
FOR i:=1 TO npar DO yar[i]:=SQRT(ABS(mat[(i-1)*nmat+i]));
FOR i:=1 TO npar DO FOR j:=1 TO npar DO
 BEGIN rr:=yar[i]*yar[j];   IF(rr>=1E-12)THEN
 mat[(i-1)*nmat+j]:=mat[(i-1)*nmat+j]/rr ELSE mat[(i-1)*nmat+j]:=1;
 IF(i<>j)THEN cmax:=AMAX1(cmax,ABS(mat[(i-1)*nmat+j]));
 IF(i<>j)THEN spu:=spu+ABS(mat[(i-1)*nmat+j]);
 END;
spu:=spu/nmat;
Writeln(fout,'Correlation matrix,  c_max= ',cmax:8:5,' sp=',spu:8:4);
FOR i:=1 TO npar DO
 BEGIN FOR j:=1 TO npar DO Write(fout,mat[(i-1)*nmat+j]:7:3);
 Writeln(fout);
 END;
END;
CONSTRUCTOR TParameter.Create(np,nm,m:INTEGER);
VAR n2,n:INTEGER;
BEGIN npar:=np; mpoint:=m; n2:=nm*nm+1; n:=npar+1;
SetLength(mat,n2); SetLength(par,n);
SetLength(yar,n); SetLength(dia,n); SetLength(ndrv,n);
END;
PROCEDURE TParameter.Arith(VAR y,w,f:REAL; i,typ:INTEGER);
BEGIN
END;
PROCEDURE TParameter.Regul(iit:INTEGER);
BEGIN
END;
PROCEDURE TestArith(VAR y,w,f:REAL; i,typ:INTEGER);
{The arithmetic procedure for least squares minimization;
    f  = a*x^b*exp(-g*x)
 computes function value + 1-st derivatives, May 11 1999}
VAR x,z:REAL;
BEGIN x:=FLOAT(i)*0.1; WITH testlsm DO
 BEGIN z:=exp(par[2].p*ln(x)-par[3].p*x);
 f:=par[1].p*z; y:=onl[i]; w:=1/(ABS(y)+1);
 par[1].drv:=z; par[2].drv:=par[1].p*ln(x)*z; par[3].drv:=-par[1].p*x*z;
 END;
END;
PROCEDURE TestRegul(iit:INTEGER);
BEGIN {Dummy}
END;
PROCEDURE TestVGFit;
{Purpose: to test gfit  Version: Jun 28 1998}
VAR n,m,i:INTEGER; c,cmax,x,z:REAL;
{---------------------}
BEGIN n:=3; {Number of parameters}
m:=100; {Number of points}
testlsm:=TParameter.Create(n,n,m);
c:=0.1;
{textfile FOUT should be opened}
WITH testlsm DO
 BEGIN FormCB; FormParameters; WITH cb DO
  BEGIN gnit:=13; ipr:=1; step:=0.75; damp:=0.01;
  END;
 {Exact parameters}
 par[1].p:=800;    par[1].pl:=1;   par[1].pu:=10000000;
 par[2].p:=0.3333; par[2].pl:=0.1; par[2].pu:=10;
 par[3].p:=0.5;    par[3].pl:=0.1; par[3].pu:=3;
 {Build quasi-experimental values}
 FOR i:=1 TO m DO
  BEGIN x:=c*i; z:=par[1].p*exp(par[2].p*ln(x)-par[3].p*x);
  onl[i]:=z+SIGN(Random-0.5)*SQRT(z);
  END;
 par[1].p:=1000;    {Initial parameters}
 par[2].p:=0.5;
 par[3].p:=0.75;
 LeastSquaresFit(TestArith,TestRegul);
 prifch; {Fitting characteristics}
 cmax:=0;
 lsmprn(TestArith,cmax); {Fitting graph; cmax= graph max; if 0,set auto}
 primat; {Correlation coefficients}
 END;
{results in FOUT; if FOUT then not needed should be closed}
testlsm.Free;
END;
{-----------------------------------------------}
BEGIN
END.
